/*
 *
 * framebuff.c
 *
 */


#define FRAMEBUFF_C

#include "framebuff.h"

char *r_graph_mem = 0;
char *v_graph_mem;
char *d_graph_mem; 
long *d_z_mem;
unsigned char d_cmap[64];


XWND XW;

/* --------------------------------- X stuff ---------------------------- */


/* ------------------------- X_SHM error handlers ----------------------- */

#ifdef X_SHM

int gXErrorFlag = 0;

int HandleXError(dpy, event)
     Display *dpy;
     XErrorEvent *event;
{
  gXErrorFlag = 1;

  return 0;
}

void InstallXErrorHandler()
{
  int HandleXError();

  XSetErrorHandler(HandleXError);
  XFlush(XW.display);
}

void DeInstallXErrorHandler()
{
  XSetErrorHandler(NULL);
  XFlush(XW.display);
}
#endif


/* ------------------------------ OpenW ---------------------------------- */

#ifdef X_SHM
int CompletionType = -1;
#endif

static void OpenW(XWND *X) 
{
  XVisualInfo vinfo;  
  XSizeHints hint;
  unsigned int fg, bg;
  int    screen;
  X->display = XOpenDisplay("");
  if (X->display == NULL) 
    {
    fprintf(stderr, "Can not open display\n");
    exit(-2);
    }

#ifdef X_SHM
  if(X->shmemFlag)
    CompletionType = XShmGetEventBase(XW.display) + ShmCompletion;
#endif

  screen = DefaultScreen (X->display);
  
  /* Fill in hint structure */

  hint.x = 200;
  hint.y = 300;
  hint.width  = D_WIDTH;
  hint.height = D_HEIGHT;
  hint.flags = PPosition | PSize;
  
  /* Get some colors */
  
  bg = WhitePixel (X->display, screen);
  fg = BlackPixel (X->display, screen);
  
  /* Make the window */
 

      if (!XMatchVisualInfo (X->display, screen, 8, PseudoColor, &vinfo) || 
          !XMatchVisualInfo (X->display, screen, 8, GrayScale,   &vinfo)) 
	{
	fprintf(stderr, "requires 8 bit display\n");
	exit(-1);
        }

    X->window = XCreateSimpleWindow (X->display,
				 DefaultRootWindow (X->display),
				 hint.x, hint.y,
				 hint.width, hint.height,
				 4, fg, bg);
  
  XSelectInput(X->display, X->window, StructureNotifyMask);

  /* Tell other applications about this window */
  
  XSetStandardProperties (X->display, 
			  X->window, 
			  "TAGL", 
			  "TAGL", None, NULL, 0, &hint);
  
  /* Map window. */

  XMapWindow(X->display, X->window);

  /* Wait for map. */
  while(1) {
    XEvent	xev;

    XNextEvent(X->display, &xev);
    if(xev.type == MapNotify && xev.xmap.event == X->window)
      break;
  }
  XSelectInput(X->display, X->window, KeyPressMask);
}



/* ------------------------------ OpenDSPL ----------------------------- */


static void OpenDSPL(XWND *X)
{
  int ncolors = 64;
  XColor xcolor;
  int i;
  int j;
  long tmp_pixel;
  XWindowAttributes xwa;

  OpenW(X);

  X->gc = XCreateGC(X->display, X->window, 0, 0);
  XGetWindowAttributes(X->display, X->window, &xwa);
  X->cmap = /* XCreateColormap(X->display, X->window, xwa.visual, AllocNone); */
    XDefaultColormap(X->display, DefaultScreen(X->display));
  XSetWindowColormap(X->display, X->window, X->cmap);

  xcolor.flags = DoRed | DoGreen | DoBlue;

  for (i=0; i<ncolors; i++) 
    {
    xcolor.red   = i << 10;
    xcolor.green = i << 10;
    xcolor.blue  = i << 10;
    XAllocColor(X->display, X->cmap, &xcolor);
    d_cmap[i] = xcolor.pixel;
    }

  X->ximage = NULL;
}

/* ------------------------------ OpenXIMG ----------------------------- */


void OpenXIMG(XWND *X)

{

  /* Allocate memory for image spaces. */

#ifdef X_SHM
  X->ximage = NULL;

  if (X->shmemFlag) 
    {
      Visual *fc_visual;
      int depth;

      X->ximage = XShmCreateImage(X->display, None, 8, ZPixmap, NULL,
				    &(X->shminfo), D_WIDTH, D_HEIGHT);

    /* If no go, then revert to normal Xlib calls. */

      if (X->ximage == NULL) 
	{
	  X->shmemFlag = 0;
	  fprintf(stderr, "Shared memory error, disabling.\n");
	  fprintf(stderr, "Ximage error.\n");
	  goto shmemerror;
	}

    /* Success here, continue. */

      X->shminfo.shmid = shmget(IPC_PRIVATE, (X->ximage->bytes_per_line *
					      X->ximage->height),
				IPC_CREAT | 0777);

      if (X->shminfo.shmid < 0) 
	{
	  XDestroyImage(X->ximage);
	  X->ximage = NULL;
	  X->shmemFlag = 0;
	  fprintf(stderr, "Shared memory error, disabling.\n");
	  fprintf(stderr, "Seg. id. error.\n");
	  goto shmemerror;
	}

      X->shminfo.shmaddr = (char *) shmat(X->shminfo.shmid, 0, 0);
      if (X->shminfo.shmaddr == ((char *) -1)) 
	{
	  XDestroyImage(X->ximage);
	  X->ximage = NULL;
	  X->shmemFlag = 0;
	  fprintf(stderr, "Shared memory error, disabling.\n");
	  fprintf(stderr, "Address error.\n");
	  goto shmemerror;
	}

      X->ximage->data = X->shminfo.shmaddr;
      X->graph_mem    = (unsigned char *) X->ximage->data;
      X->shminfo.readOnly = False;

      XShmAttach(X->display, &(X->shminfo));
      XSync(X->display, False);

      if (gXErrorFlag) 
	{
	  /* Ultimate failure here. */
	  XDestroyImage(X->ximage);
	  X->ximage = NULL;
	  X->shmemFlag = 0;
	  fprintf(stderr, "Shared memory error, disabling.\n");
	  gXErrorFlag = 0;
	  goto shmemerror;
	} 
      else 
	shmctl(X->shminfo.shmid, IPC_RMID, 0);

      fprintf(stderr, "Sharing memory.\n");
  }
  else

#endif
    
    {
shmemerror:
    X->graph_mem = (char *) malloc(D_WIDTH * D_HEIGHT);
    }
}

/* ------------------------- CloseXIMG ------------------------------ */

void CloseXIMG(XWND *X)
{

#ifdef X_SHM
  if (X->ximage != NULL) 
    {
    XShmDetach(X->display, &(X->shminfo));
    XDestroyImage(X->ximage);
    shmdt(X->shminfo.shmaddr);
    X->ximage    = NULL;
    X->graph_mem = NULL;
  }
#endif

  if (X->graph_mem != NULL) 
    free(X->graph_mem);
}


/* ------------------------- DisplayXIMG ------------------------------ */

void DisplayXWND(XWND *X)

{
  char dummy;
  Visual *FindFullColorVisual();
  Visual *fc_visual;
  int depth;

  if (X->ximage == NULL) 
    {
    X->ximage = XCreateImage(X->display, None, 8, ZPixmap, 0, &dummy,
			    D_WIDTH  * 16,
			    D_HEIGHT * 16, 8, 0);
    }

#ifdef X_SHM
    if (X->shmemFlag) 
      {
	XShmPutImage(X->display, X->window, X->gc, 
		     X->ximage, 
		     0, 0, 0, 0,
		     D_WIDTH, D_HEIGHT, True);
	XFlush(X->display);

/*      
	while(1) 
	  {
	    XEvent xev;
	    XNextEvent(X->display, &xev);
	    if(xev.type == CompletionType)
	      break;
	  }
*/

      }
    else 
#endif
      {
	X->ximage->data = X->graph_mem; 
	XPutImage(X->display, X->window, X->gc, X->ximage, 
		  0, 0, 0, 0, X->ximage->width, X->ximage->height);
      }
  
}




/* --------------------------------------------------------------------- */


void OpenS()
   {

     printf("Opening TAGL screen\n");
     OpenDSPL(&XW);
#ifdef X_SHM
     XW.shmemFlag = XShmQueryExtension(XW.display);
     printf("SHM flag = %d\n",XW.shmemFlag);
#endif
     printf("Creating XImage structure\n");
     OpenXIMG(&XW);
     printf("TAGL Screen created\n");
     d_z_mem     = NEWARRAY(long, D_WIDTH * D_HEIGHT);
     r_graph_mem = v_graph_mem = d_graph_mem = XW.graph_mem;
   }

void CloseS()
   {
     CloseXIMG(&XW);
   }

void graph_swap()
   {
     DisplayXWND(&XW);
   }



int readkbd(char *c)
{
	char	q;
	XEvent event;

	if (!XCheckWindowEvent(XW.display, XW.window, KeyPressMask, &event))
	    return(0);

	if (event.type == KeyPress)
		if (XLookupString((XKeyEvent *)&event, &q, 1, NULL, NULL) > 0)
		  {
		    *c = q;
		    return 1;
		  }
	
	return(0);
}


int gl_locator(wx, wy)
	int	*wx, *wy;
{
	Window		rootw, childw;
	int		x, y;
	unsigned int	mask;

	XQueryPointer(XW.display, XW.window, &rootw, &childw, &x, &y, wx, wy, 
		      &mask);

	*wy = D_HEIGHT - *wy;

	return(mask >> 8);
}

