/*
 *
 * gl_3d.c
 *
 */

#include <math.h>
#include "gl_3d.h"

int Zoom     = 16;
int ZoomStep = 4;

Vector _Light = {0, 0, -GL_BIG};
int    _ShadeMode = 0;

int _S_z1 = -GL_BIG;
int _S_z2 =  GL_BIG;

void Normalize(Vector *V)
{
     VectorF W;
     dbl k;

     SetVector(&W, V->x, V->y, V->z);

     k = NORM2(&W);
     k = k ? 16384.0 / sqrt(k) : 0.0;

     ScaleVector(&W,k);
     SetVector(V,W.x,W.y,W.z);
}

/*
 * -------------------------------------------------------------------
 * shading
 * -------------------------------------------------------------------
 */


void SetDepthQRange(int z1, int z2)
{
_S_z1 = z1;
_S_z2 = z2;
}

void GetDepthQRange(int *z1, int *z2)
{
*z1 = _S_z1;
*z2 = _S_z2;
}

void SetShadeMode(int mode)
{
_ShadeMode = mode;
};

int GetShadeMode()
{
return _ShadeMode;
}


void SetLight(Vector *v)
{
_Light = *v;
}

/*
 * -------------------------------------------------------------------
 * zoom
 * -------------------------------------------------------------------
 */


void SetZoom(z)
     int z;
{
  Zoom = RANGE(z,1,100);
}

int GetZoom()
{
  return Zoom;
}

void SetZoomStep(zs)
     int zs;
{
  ZoomStep = RANGE(zs,1,10);
}

int GetZoomStep()
{
  return ZoomStep;
}


void ZoomIn()
{
  Zoom += ZoomStep;
  Zoom = MIN(Zoom,100);
}

void ZoomOut()
{
  Zoom -= ZoomStep;
  Zoom = MAX(Zoom,1);
}


/*
 * -------------------------------------------------------------------
 * Rot vector
 * -------------------------------------------------------------------
 */


void RotX(r, v)

     int r;
     Vector *v;

   {
   long t1 = v->y;
   long t2 = v->z;
   long s = SIN(r),c = COS(r);

   v->y = ( c * t1 + s * t2) >> SIN_SHIFT;
   v->z = (-s * t1 + c * t2) >> SIN_SHIFT;
   }

void RotY(r, v)

     int r;
     Vector *v;

   {
   long t1 = v->x;
   long t2 = v->z;
   long s = SIN(r),c = COS(r);

   v->x = ( c * t1 + s * t2) >> SIN_SHIFT;
   v->z = (-s * t1 + c * t2) >> SIN_SHIFT;
   }
   
void RotZ(r, v)

     int r;
     Vector *v;

   {
   long t1 = v->x;
   long t2 = v->y;
   long s = SIN(r),c = COS(r);

   v->x = ( c * t1 + s * t2) >> SIN_SHIFT;
   v->y = (-s * t1 + c * t2) >> SIN_SHIFT;
   }

void Rot(v, axis, r)

   Vector *v;
   char axis;
   int r;

   {
   long t1, t2;
   long s = SIN(r), c = COS(r);
   switch(axis)
      {
      	case 'x':
      	case 'X':
      	   t1 = v->y;
      	   t2 = v->z;
	   v->y = ( c * t1 + s * t2) >> SIN_SHIFT;
   	   v->z = (-s * t1 + c * t2) >> SIN_SHIFT;
      	   break;
      	case 'y':
      	case 'Y':
      	   t1 = v->x;
      	   t2 = v->z;
	   v->x = ( c * t1 + s * t2) >> SIN_SHIFT;
   	   v->z = (-s * t1 + c * t2) >> SIN_SHIFT;      	
      	   break;
      	case 'z':
      	case 'Z':
      	   t1 = v->x;
      	   t2 = v->y;
	   v->x = ( c * t1 + s * t2) >> SIN_SHIFT;
   	   v->y = (-s * t1 + c * t2) >> SIN_SHIFT;
      	   break;
	default:
	   fprintf(stderr,"rot: invalid axis\n");
	   exit(1);
	   break;      	   
      }
   }

   
   
void Project(V, P)
     
     Vector *V;
     Pixel  *P;

   {
   long x = V->x;
   long y = V->y;
   long z = V->z;
/*   long YE = ((49152 - y) >> 9); */

   long YE = 65536 >> 9;
   long X,Y;

#ifdef D_HIRES
   X = (D_WIDTH >> 1)  + (((((x*YE*3) >> 14)) * Zoom) >> 4);
   Y = (D_HEIGHT >> 1) + (((((y*YE*3) >> 14)) * Zoom) >> 4);
#else
   X = (D_WIDTH >> 1)  + (((((x*YE*3) >> 15)) * Zoom) >> 4);
   Y = (D_HEIGHT >> 1) + (((((y*YE*3) >> 15)) * Zoom) >> 4);
#endif

/*   z += GL_BIG; */

   P->x = X;
   P->y = Y;
   P->z = z;

   }


/*
 * -------------------------------------------------------------------
 * Draw
 * -------------------------------------------------------------------
 */

void DrawMesh(m,draw)

   Mesh *m; 
   polyfun draw;

   {
   int i,j,id1,id2;

   int shadevertex;
   int shadeface; 
   long c;


  if(!(m->mode & M_SMOOTH)    &&
      (_ShadeMode == S_LIGHT) && 
      (draw == P_SHADE_MODE || draw == P_SHADE_DITHER_MODE))
    draw = P_FILL_COLOR_MODE;


  if(!(m->mode & M_SMOOTH)    &&
      (_ShadeMode == S_LIGHT) && 
      (draw == P_ZBUFF_SHADE_MODE))
    draw = P_ZBUFF_MODE;


  shadevertex = _ShadeMode ? 
                      ((draw == P_SHADE_MODE) || 
                       (draw == P_SHADE_DITHER_MODE) ||
		       (draw == P_ZBUFF_SHADE_MODE))
                    : 0;

  shadeface   = (_ShadeMode == S_LIGHT) ? !shadevertex : 0;


  for(i=0; i<m->nvertex; i++)
     Project(&(m->vertex[i]),&(m->proj[i]));

  if(!(m->mode & M_CONVEX))    
   for(i=0; i<m->nfaces; i++)
      {
      if((m->mode & M_KAOS) || (m->face[i].N.z > 0))     
	 {
	 _poly1.npv2 = 0;
	 for(j=0; j<m->face[i].npts; j++)
	    {
	    v2_push(&_poly1, &(m->proj[m->face[i].pts[j]]));
	    if(shadevertex)
	      {
	      switch(_ShadeMode)
		{
		case S_DEPTHQ:
		  c = _poly1.pv2[_poly1.npv2 - 1]->z;
		  c = _S_z2 - c;
		  c = c < 0 ? 0 : c << 6;
		  c /= (_S_z2 - _S_z1);
		  c = c > 63 ? 63 : c;  
		  break;
		case S_LIGHT:
		  c = m->N[m->face[i].pts[j]].x * _Light.x +
		      m->N[m->face[i].pts[j]].y * _Light.y +
		      m->N[m->face[i].pts[j]].z * _Light.z ;
		  c = c < 0 ? -c : 0;
		  c = c >> 22;
		  c = c > 63 ? 63 : c;
		  break;
		}
	      _poly1.pv2[_poly1.npv2 - 1]->c = c;
	      }
	  }
	 if(shadeface)
	   {
 	   c = m->face[i].N.x * _Light.x +
	       m->face[i].N.y * _Light.y +
	       m->face[i].N.z * _Light.z ;
	   c = c < 0 ? -c : 0;
	   c = c >> 22;
	   c = c > 63 ? 63 : c;
	   }
	 _do_clip();
	 if(shadeface)
	    _poly1.pv2[0]->c = c;
	 draw();
         }
      }

   if(!(m->mode & M_KAOS))   
   for(i=0; i<m->nfaces; i++)
      {
        if(m->face[i].N.z <= 0)    
	 {
	 _poly1.npv2 = 0;	 	
	 for(j=0; j<m->face[i].npts; j++)
	    {
	    v2_push(&_poly1, &(m->proj[m->face[i].pts[j]]));
	    if(shadevertex)
	      {
	      switch(_ShadeMode)
		{
		case S_DEPTHQ:
		  c = _poly1.pv2[_poly1.npv2 - 1]->z;
		  c = _S_z2 - c;
		  c = c < 0 ? 0 : c << 6;
		  c /= (_S_z2 - _S_z1);
		  c = c > 63 ? 63 : c;  
		  break;
		case S_LIGHT:
		  c = m->N[m->face[i].pts[j]].x * _Light.x +
		      m->N[m->face[i].pts[j]].y * _Light.y +
		      m->N[m->face[i].pts[j]].z * _Light.z ;
		  c = c > 0 ? c : 0;
		  c = c >> 22;
		  c = c > 63 ? 63 : c;
		  break;
		}
	      _poly1.pv2[_poly1.npv2 - 1]->c = c;
	      }
	    }

	 if(shadeface)
	   {
 	   c = m->face[i].N.x * _Light.x +
	       m->face[i].N.y * _Light.y +
	       m->face[i].N.z * _Light.z ;
	   c = c > 0 ? c : 0;
	   c = c >> 22;
	   c = c > 63 ? 63 : c;
	   }
	 _do_clip();
	 if(shadeface)
	    _poly1.pv2[0]->c = c;
	 draw();
         }
      }
   }


void DrawLink(l,draw)

     Link *l;
     polyfun draw;

   {
   if(l->N.z < 0)
      {
      DrawNode(l->n2, draw);
      DrawNode(l->n1, draw);
      }
   else
      {
      DrawNode(l->n1, draw);
      DrawNode(l->n2, draw);
      }
   }

void DrawNode(n,draw)

   Node *n;
   polyfun draw;

   {
   if(n->m)
      DrawMesh(n->m, draw);
   if(n->l)
      DrawLink(n->l, draw);   
   }

/*
 * -------------------------------------------------------------------
 * Translate
 * -------------------------------------------------------------------
 */

void TranslateMesh(m, v)

   Mesh *m;
   Vector *v;

{
  int i;
  for(i=0; i<m->nvertex; i++)
    {
    m->vertex[i].x += v->x;
    m->vertex[i].y += v->y;
    m->vertex[i].z += v->z;
    }
}

void TranslateLink(l,v)

   Link *l;
   Vector *v;

   {
   TranslateNode(l->n1, v);
   TranslateNode(l->n2, v);
   }

void TranslateNode(n,v)

   Node   *n;
   Vector *v;

   {
   if(n->l)
   	TranslateLink(n->l, v);
   if(n->m)
	TranslateMesh(n->m, v);	
   }

/*
 * -------------------------------------------------------------------
 * Rot
 * -------------------------------------------------------------------
 */

void RotMesh(m,axis,r)

   Mesh *m;
   char axis;
   int r;

   {
   int i;
   for(i=0; i<m->nfaces; i++)
   	Rot(&(m->face[i].N),axis,r);
   for(i=0; i<m->nvertex; i++)
   	Rot(&(m->vertex[i]),axis,r);
   if(m->mode & M_SMOOTH)
   	for(i=0; i<m->nvertex; i++)
   	   Rot(&(m->N[i]),axis,r);
   }

void RotLink(l,axis,r)

   Link *l;
   char axis;
   int r;

   {
   RotNode(l->n1, axis, r);
   RotNode(l->n2, axis, r);
   Rot(&(l->N), axis, r);
   }

void RotNode(n,axis,r)

   Node *n;
   char axis;
   int r;

   {
   if(n->l)
   	RotLink(n->l, axis, r);
   if(n->m)
	RotMesh(n->m, axis, r);	
   }


/*
 * -------------------------------------------------------------------
 * Copy
 * -------------------------------------------------------------------
 */


Mesh *MeshCpy(Mesh *m)
   {
   Mesh *result = NEW(Mesh);

   result->nvertex = m->nvertex;
   result->vertex = NEWARRAY(Vector,m->nvertex);
   memcpy(result->vertex, m->vertex, sizeof(Vector) * m->nvertex);

   result->mode = m->mode;
   if(result->mode & M_SMOOTH)
	{
	result->N = NEWARRAY(Vector,m->nvertex);
	memcpy(result->N, m->N, sizeof(Vector) * m->nvertex);
	}

   result->proj = NEWARRAY(Pixel, m->nvertex);
   memcpy(result->proj, m->proj, sizeof(Pixel) * m->nvertex);

   result->nfaces = m->nfaces;
   result->face = NEWARRAY(Face, m->nfaces);
   memcpy(result->face, m->face, sizeof(Face)*m->nfaces);

   return result;
   }

Node *NodeCpy(Node *n)
   {
   Node *result = NEW(Node);
   result->l = n->l ? LinkCpy(n->l) : NULL;
   result->m = n->m ? MeshCpy(n->m) : NULL;
   return result;	
   }


Link *LinkCpy(Link *l)
   {
   Link *result = NEW(Link);
   result->n1 = NodeCpy(l->n1);
   result->n2 = NodeCpy(l->n2);
   result->N  = l->N;
   return result;
   }


/*
 * -------------------------------------------------------------------
 * Create - build
 * -------------------------------------------------------------------
 */


Node *Mesh2Node(Mesh *m)
   {
   Node *result = NEW(Node);
   result->l = NULL;
   result->m = m;
   return result;
   }

Node *Link2Node(Link *l)
   {
   Node *result = NEW(Node);
   result->l = l;
   result->m = NULL;
   return result;
   }

Link *MakeLink(Node *n1, Node *n2, Vector *N)
   {
   Link *result = NEW(Link);
   result->n1 = n1;
   result->n2 = n2;
   result->N  = *N;
   return result;
   }

Mesh *MakeBox(Vector *v1, Vector *v2, int convex)
   {
   int x1 = v1->x;
   int y1 = v1->y;
   int z1 = v1->z;

   int x2 = v2->x;
   int y2 = v2->y;
   int z2 = v2->z;

   Mesh *result = NEW(Mesh);
   result->nvertex = 8;
   result->vertex = NEWARRAY(Vector, result->nvertex);
   result->mode   = convex ? M_CONVEX : M_NONE;
   result->proj   = NEWARRAY(Pixel, result->nvertex);
   result->nfaces = convex ? 6 : 4;
   result->face   = NEWARRAY(Face, result->nfaces);

   SetVector(&(result->vertex[0]),x1,y1,z1);
   SetVector(&(result->vertex[1]),x1,y1,z2);
   SetVector(&(result->vertex[2]),x1,y2,z1);
   SetVector(&(result->vertex[3]),x1,y2,z2);
   SetVector(&(result->vertex[4]),x2,y1,z1);
   SetVector(&(result->vertex[5]),x2,y1,z2);
   SetVector(&(result->vertex[6]),x2,y2,z1);
   SetVector(&(result->vertex[7]),x2,y2,z2);

   SetFace4(&(result->face[0]), 4, 0, 1, 5);
   SetFace4(&(result->face[1]), 4, 5, 7, 6);
   SetFace4(&(result->face[2]), 3, 2, 6, 7);
   SetFace4(&(result->face[3]), 1, 0, 2, 3);

   if(convex)
     {
     SetFace4(&(result->face[4]), 2, 0, 4, 6);
     SetFace4(&(result->face[5]), 5, 1, 3, 7);
     }


   ComputeNormals(result);

   return result;
   }


Mesh* MakeCyl(int r, int h, int prec, int convex)
{
  int nstep = 1 << prec;
  Mesh *result; 
  int i,j;
  long s,c;
  int idx;
  int zz = h >> 1;
  long x,y;

  if(nstep < 4 || nstep > 256)
    {
      fprintf(stderr,"MakeCyl: invalid precison: %d\n",prec);
      exit(1);
    }

  result = NEW(Mesh);
  result->nfaces  = convex ? nstep + 2 : nstep;
  result->nvertex = 2 * nstep;
  result->vertex = NEWARRAY(Vector, result->nvertex);
  result->face   = NEWARRAY(Face, result->nfaces); 
  result->mode   = convex ? M_CONVEX : M_NONE;
  result->proj   = NEWARRAY(Pixel, result->nvertex);


  for(i=0; i<nstep; i++)
    {
      idx = i << (8 - prec);
      s = SIN(idx);
      c = COS(idx);
      x = (s * r) >> SIN_SHIFT;
      y = (c * r) >> SIN_SHIFT;

      j = i+1;
      j = (j == nstep) ? 0 : j;

      SetVector(&(result->vertex[2*i  ]),x,y, zz);
      SetVector(&(result->vertex[2*i+1]),x,y,-zz);

      SetFace4(&(result->face[i]),2*i,2*i+1,2*j+1,2*j); 

      if(convex)
	{
	  result->face[nstep  ].pts[i]          = 2*i;
	  result->face[nstep+1].pts[nstep-i-1]  = 2*i+1;
	}
    }

  if(convex)
    {
      result->face[nstep  ].npts = nstep;
      result->face[nstep+1].npts = nstep;
    }

  ComputeNormals(result);

  return result;
}


Mesh* MakeCone(int r1, int r2, int h, int prec, int convex)
{
  int nstep = 1 << prec;
  Mesh *result; 
  int i,j;
  long s,c;
  int idx;
  int zz = h >> 1;
  long x1,y1;
  long x2,y2;

  if(nstep < 4 || nstep > 256)
    {
      fprintf(stderr,"MakeCyl: invalid precison: %d\n",prec);
      exit(1);
    }

  result = NEW(Mesh);
  result->nfaces  = convex ? nstep + 2 : nstep;
  result->nvertex = 2 * nstep;
  result->vertex = NEWARRAY(Vector, result->nvertex);
  result->face   = NEWARRAY(Face, result->nfaces); 
  result->mode   = convex ? M_CONVEX : M_NONE;
  result->proj   = NEWARRAY(Pixel, result->nvertex);


  for(i=0; i<nstep; i++)
    {
      idx = i << (8 - prec);
      s = SIN(idx);
      c = COS(idx);
      x1 = (s * r1) >> SIN_SHIFT;
      y1 = (c * r1) >> SIN_SHIFT;
      x2 = (s * r2) >> SIN_SHIFT;
      y2 = (c * r2) >> SIN_SHIFT;

      j = i+1;
      j = (j == nstep) ? 0 : j;

      SetVector(&(result->vertex[2*i  ]),x2,y2, zz);
      SetVector(&(result->vertex[2*i+1]),x1,y1,-zz);

      SetFace4(&(result->face[i]),2*i,2*i+1,2*j+1,2*j); 

      if(convex)
	{
	  result->face[nstep  ].pts[i]          = 2*i;
	  result->face[nstep+1].pts[nstep-i-1]  = 2*i+1;
	}
    }

  if(convex)
    {
      result->face[nstep  ].npts = nstep;
      result->face[nstep+1].npts = nstep;
    }

  ComputeNormals(result);

  return result;
}



/*
 * -------------------------------------------------------------------
 * Normals stuff
 * -------------------------------------------------------------------
 */

void ComputeNormals(Mesh *m)
   {
   int i; 
   VectorF A,B,N;
   dbl k;

   for(i=0; i<m->nfaces; i++)
     {
     A.x = m->vertex[m->face[i].pts[0]].x - 
           m->vertex[m->face[i].pts[1]].x;  
     A.y = m->vertex[m->face[i].pts[0]].y - 
           m->vertex[m->face[i].pts[1]].y;  
     A.z = m->vertex[m->face[i].pts[0]].z - 
           m->vertex[m->face[i].pts[1]].z;  

     B.x = m->vertex[m->face[i].pts[2]].x - 
           m->vertex[m->face[i].pts[1]].x;  
     B.y = m->vertex[m->face[i].pts[2]].y - 
           m->vertex[m->face[i].pts[1]].y;  
     B.z = m->vertex[m->face[i].pts[2]].z - 
           m->vertex[m->face[i].pts[1]].z;  

     SetVector(&N,
	       A.y*B.z - A.z*B.y,
	       A.z*B.x - A.x*B.z,
	       A.x*B.y - A.y*B.x
	      );

     k = NORM2(&N);
     k = k ? 16384.0 / sqrt(k) : 0.0;

     ScaleVector(&N,k);
     SetVector(&(m->face[i].N),N.x,N.y,N.z);
     }
   }


int InFace(int idx, Face *f)
   {
   int i;
   for(i=0; i<f->npts; i++)
     if(f->pts[i] == idx)
       return 1;
   return 0;
   }

void SmoothMesh(Mesh *m)
   {
   int i,j;
   VectorF N;
   dbl k;

   m->N = NEWARRAY(Vector, m->nvertex);

   for(i=0; i<m->nvertex; i++)
     {
       SetVector(&N,0.0,0.0,0.0);

       for(j=0; j<m->nfaces; j++)
	 {
	   if(InFace(i,&(m->face[j])))
	     {
	       N.x += m->face[j].N.x;
	       N.y += m->face[j].N.y;
	       N.z += m->face[j].N.z;
	     }
	 }

       k = NORM2(&N);
       k = k ? 16384.0 / sqrt(k) : 0.0;

       ScaleVector(&N,k);
       SetVector(&(m->N[i]),N.x,N.y,N.z);
     }
   m->mode |= M_SMOOTH;

   }




