/*
 *
 * gl_3d.h
 *
 */

 /*
  * This modules provides facilities for
  * modelling and viewing Linear Separable Trees.
  *
  */


#ifndef GL_3D_H
#define GL_3D_H

#include "gl_2d.h"
#include "gl_sintab.h" 

#define GL_BIG 16384

#define MAX_PT_PER_FACE 512

#define M_NONE   0
#define M_CONVEX 1
#define M_SMOOTH 2
#define M_KAOS   4

#define S_NONE   0
#define S_LIGHT  1
#define S_DEPTHQ 2

typedef struct
   {
   int x,y,z;
   } Vector;

typedef struct
   {
   dbl x,y,z;
   } VectorF;

#define SetVector(v,xx,yy,zz) {               \
			      (v)->x = (xx);  \
			      (v)->y = (yy);  \
			      (v)->z = (zz);  \
                              }

#ifndef SQR
#define SQR(x) ((x)*(x))
#endif

#define NORM2(V) (SQR((V)->x)+SQR((V)->y)+SQR((V)->z))

#define ScaleVector(v,k) {             \
			 (v)->x *= k;  \
 			 (v)->y *= k;  \
			 (v)->z *= k;  \
                         } 

typedef struct
   {
   int npts;
   int pts[MAX_PT_PER_FACE];
   Vector N;
   } Face;

#define SetFace3(F,a,b,c)  {                     \
			   (F)->npts = 3;        \
			   (F)->pts[0] = a;      \
			   (F)->pts[1] = b;      \
			   (F)->pts[2] = c;      \
                           }                


#define SetFace4(F,a,b,c,d)  {                     \
			     (F)->npts = 4;        \
			     (F)->pts[0] = a;      \
			     (F)->pts[1] = b;      \
			     (F)->pts[2] = c;      \
			     (F)->pts[3] = d;      \
                             }                

typedef struct mesh
   {
   int nvertex;
   Vector *vertex;
   Vector *N;
   Pixel  *proj;
   int nfaces;
   Face *face;
   int mode;
   } Mesh;

struct link;

typedef struct
   {
   struct link *l;
   struct mesh *m;
   } Node;
   
typedef struct link
   {
   Node *n1,*n2;
   Vector N;
   } Link;
   

void Normalize       PRM((Vector *V));

void SetShadeMode    PRM((int mode));
int  GetShadeMode    PRM(());
void SetLight        PRM((Vector *L));
void SetDepthQRange  PRM((int  z1, int  z2));
void GetDepthQRange  PRM((int *z1, int *z2));

void SetZoom     PRM((int   z));
int  GetZoom     PRM(());
void SetZoomStep PRM((int  zs));
int  GetZoomStep PRM(());
void ZoomIn      PRM(());
void ZoomOut     PRM(());


void RotX     PRM((int r, Vector *v));
void RotY     PRM((int r, Vector *v));
void RotZ     PRM((int r, Vector *v));
void Rot      PRM((Vector *v, char axis, int r));

void Project  PRM((Vector *V, Pixel *P));

void DrawMesh  PRM((Mesh *m, polyfun draw));
void DrawLink  PRM((Link *l, polyfun draw));
void DrawNode  PRM((Node *n, polyfun draw));

void RotMesh   PRM((Mesh *m, char axis, int r));
void RotLink   PRM((Link *l, char axis, int r));
void RotNode   PRM((Node *n, char axis, int r));

void TranslateMesh PRM((Mesh *m, Vector *v));
void TranslateLink PRM((Link *l, Vector *v));
void TranslateNode PRM((Node *n, Vector *v));

Mesh *MeshCpy  PRM((Mesh *m));
Node *NodeCpy  PRM((Node *n));
Link *LinkCpy  PRM((Link *l));

Node *Mesh2Node PRM((Mesh *m));
Node *Link2Node PRM((Link *l));
Link *MakeLink  PRM((Node *n1, Node *n2, Vector *N));
Mesh *MakeBox   PRM((Vector *v1, Vector *v2, int convex));
Mesh *MakeCyl   PRM((int r,          int h, int prec, int convex));
Mesh *MakeCone  PRM((int r1, int r2, int h, int prec, int convex));

void  ComputeNormals PRM((Mesh *m));
void  SmoothMesh     PRM((Mesh *m));

#endif
