
/*
 *
 * rotate.c
 * rotate 3D objects from 
 * standard file format.
 * This file is part of the TAGL package.
 *
 */

#include "framebuff.h"
#include "gl_clip2d.h"
#include "gl_3d.h"
#include "gl_obj.h"
#include <time.h>

#define now() time(NULL)

Mesh *TheMesh;

typedef Mesh *(*readfun)();

typedef struct
{
  char    *label;
  readfun  readobj;
} Format;


Mesh *read_error PRM((char *FileName));


Vector T1 = {-1000,0,0};
Vector T2 = { 1000,0,0};
Vector T3 = {0,-1000,0};
Vector T4 = {0, 1000,0};



Format format[] =
{
  {"3DG1", gl_read_3DG1},
  {"geo",  gl_read_geo },
  {NULL,   read_error  }
};


int help   = 0;
long nframe = 0;
int tinit;

int BD = 20;
int T  = 10;



void InitMouse()
{
/*
   mouse_init("/dev/mouse", MOUSE_MOUSESYSTEMS, MOUSE_DEFAULTSAMPLERATE);
   mouse_setxrange(-10000, 10000);
   mouse_setyrange(-10000, 10000);
   mouse_setposition(D_WIDTH >> 1,D_HEIGHT >> 1);
   mouse_setwrap(MOUSE_NOWRAP);
*/
}

void InvertNormals(Mesh *m)
{
  int i;
  for(i=0; i<m->nvertex; i++)
    ScaleVector(&(m->N[i]),-1);

  for(i=0; i<m->nfaces; i++)
    ScaleVector(&(m->face[i].N),-1);
}

int gl_valuator(wx, wy)
int *wx, *wy;
{
int gl_locator PRM((int *, int *));

return gl_locator(wx,wy);
}


Mesh *read_error(char *FileName)
{
  fprintf(stderr,"rotate: invalid file format for %s\n",FileName);
  fprintf(stderr,"rotate: -format= -3DG1 or -geo\n");
  exit(1);
}


Mesh *ReadObject(char *FileName, char *FileFormat)
{
  int i;
  readfun readfile = NULL;
  for(i=0; !readfile; i++)
    {
      if(!format[i].label || !strcmp(format[i].label,FileFormat))
	readfile = format[i].readobj;
    }
  return TheMesh = readfile(FileName,10000);
} 

void InitColormap()
{
int i;
/*
for(i=0; i<64; i++)
   SetRGB(i,i,i,i);
*/
}

void Help(x,y)
   {
   Rect r,rr;
   char stats[300];
   long elapsed = now() - tinit;

   elapsed = elapsed ? elapsed : 1;

   r.x1 = x-20;       r.y1 = y-20;
   r.x2 = r.x1 + 250; r.y2 = r.y1+140;
   rr = Clip;
   Clip = r;

   sprintf(stats,"f %5d t %4d ft %3d",nframe, elapsed, 10*nframe/elapsed);

   gl_string("TAGL 3D object viewer ",x,y    ,63);
   gl_string("rotate X         1-4-7",x,y+20 ,63);
   gl_string("rotate Y         2-5-8",x,y+30 ,63);
   gl_string("rotate Z         3-6-9",x,y+40 ,63);
   gl_string("stop             0    ",x,y+50 ,63);
   gl_string("toggle shading   w    ",x,y+60 ,63);
   gl_string("zoom in-out      + -  ",x,y+70 ,63);
   gl_string("toggle help      h    ",x,y+80 ,63);
   gl_string("quit             q    ",x,y+90 ,63);
   gl_string("reset counters   r    ",x,y+100,63);
   gl_string(stats,x,y+110,63);

   DrawRect(&r,63);

   Clip = rr;
   }

void RotateObj(Node *n)
   {

   static int ds = 0;
   Vector L;
   int light = 0;
   char q = ' ';
   int rx  = 0, ry  = 0, rz  = 0;
   int rrx = 1, rry = 1, rrz = 1;
   int oldx,  oldy;
   int newx,  newy;
   polyfun p;
   int opaquemove = 0;
   int moved = 0;

   help = 1;

   gl_valuator(&oldx, &oldy);


   SetVector(&L, -GL_BIG, 0, -GL_BIG);
   Normalize(&L);
   SetLight(&L);

   SetDepthQRange(-7000,1000);

   while(q != 'q')
     {
     if( readkbd(&q) )
       switch(q)
	 {
	 case 'u': InvertNormals(TheMesh); break;  
      	 case 'a': BD -= 5; break;
      	 case 'z': BD += 5; break;
	 case '0': rx = ry = rz = 0; gl_valuator(&oldx, &oldy); break;
	 case '7': rx -= rrx; break;
	 case '4': rx = 0;    break;
	 case '1': rx += rrx; break;
	 case '8': ry -= rry; break;
	 case '5': ry = 0;    break;
	 case '2': ry += rry; break;
	 case '9': rz -= rrz; break;
	 case '6': rz = 0;    break;
	 case '3': rz += rrz; break;
	 case '+': ZoomIn();  break;
	 case '-': ZoomOut(); break;
	 case 'h': help   = help   ? 0 : 1; break;
	 case 'r': tinit = now(); nframe = 0; break;
	 case 'w': light++; light %= 4; break;
	 case 'j': TranslateNode(n,&T1); break;
	 case 'k': TranslateNode(n,&T2); break;
	 case 'i': TranslateNode(n,&T3); break;
	 case ',': TranslateNode(n,&T4); break;
	 case '!': opaquemove = opaquemove ? 0 : 1; break;
	 }

     moved = 0;

     if(rx || ry || rz)
       {
	 RotNode(n,'x',rx);
	 RotNode(n,'y',ry);
	 RotNode(n,'z',rz);
       }
     else
       {
	 int nnewx, nnewy;
	 int btn = gl_valuator(&nnewx,&nnewy);

	 newx = btn & 2 ? nnewx : oldx;
	 newy = btn & 2 ? nnewy : oldy;

	 moved =               (btn || (ABS(oldx - newx) > 20 || 
					ABS(oldy - newy) > 20)); 
/*	 moved = (btn & 2); */

	 if(btn & (1 | 4))
	   {
	   RotNode(n, 'z', (oldx - newx) >> 2);
	   }
	 else
	   {
	   RotNode(n, 'y', (oldx - newx) >> 2);
	   RotNode(n, 'x', (newy - oldy) >> 2);
           }

	 oldx = nnewx;
	 oldy = nnewy;
	 if((btn == 1) || (btn == (1 | 2)))
	   ZoomIn();
	 if((btn == 4) || (btn == (4 | 2)))
	   ZoomOut();
       }

     switch(light)
       {
       case 0:
	 SetShadeMode(S_LIGHT); 
	 p = P_ZBUFF_SHADE_MODE;
	 break;
       case 1:
	 SetShadeMode(S_DEPTHQ);
	 p = P_ZBUFF_SHADE_MODE;
	 break;
       case 2:
	 SetShadeMode(S_LIGHT);   
	 p = P_ZBUFF_MODE;
	 break;
       case 3:
	 SetShadeMode(S_NONE);
	 p = P_DRAW_MODE;
	 break;
       }

     SetClip(BD,BD,D_WIDTH-BD,D_HEIGHT-BD);
     SetClipMode(CM_COLOR | CM_Z);
     graph_clear(6);
     z_clear();
     DrawRect(&Clip,32);

     if(moved && !opaquemove)
       {
       SetShadeMode(S_NONE);
       DrawNode(n,P_DRAW_MODE);
       }
     else
       DrawNode(n,p);

     if(help)
       Help(60,50);

     graph_swap();
     nframe++;
     }

   }

int main(argc, argv)
     int    argc;
     char **argv;
   {
   Mesh *m;
   Vector V = {0,0,0};


   if(argc != 3)
     {
       fprintf(stderr,"rotate: invalid # of args\n");
       fprintf(stderr,"rotate: rotate -format filename\n");
       fprintf(stderr,"rotate: -format= -3DG1 or -geo\n");
       exit(1);
     }
   OpenS();
   InitMouse();
   InitColormap();
   SetClipMode(CM_COLOR | CM_Z);
   d_graph_mem = v_graph_mem;
   tinit = now();
   m = ReadObject(argv[2],argv[1]+1);
   SmoothMesh(m);
   RotateObj(Mesh2Node(m));
   CloseS();
   return 0;
   }
